# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "PersiaML API Documentation"
copyright = "2021, Kuaishou AI Platform"
author = "Kuaishou AI Platform"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "autoapi.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]


napoleon_numpy_docstring = True
autoapi_python_class_content = "both"
autodoc_typehints = "description"
autoapi_type = "python"
autoapi_dirs = ["../persia"]
autoapi_root = "autoapi"
autoapi_template_dir = "_autoapi_templates"
autoapi_ignore = [
    # "*/bagua/autotune/*",
]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "imported-members",
]
autoapi_member_order = "groupwise"

master_doc = "autoapi/index"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "show_powered_by": False,
    "github_user": "PersiaML",
    "github_repo": "PersiaML",
    "github_banner": True,
    "show_related": False,
    "note_bg": "#FFF59C",
}

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True


_ignore_methods = [
    # "bagua.torch_api.contrib.LoadBalancingDistributedSampler.shuffle_chunks",
]

_ignore_functions = []

_ignore_classes = []

_ignore_module = [
    "persia.version",
    "persia.sparse.emb",
    "persia.launcher",
    "persia.error",
    "persia.logger",
    "persia.prelude",
    "persia.sparse.__init__",
]


def skip_methods(app, what, name, obj, skip, options):
    if what == "method" and name in _ignore_methods:
        skip = True
        return skip

    if what == "function" and name in _ignore_functions:
        skip = True
        return skip

    if what == "class" and name in _ignore_classes:
        skip = True
        return skip

    if what == "module" and name in _ignore_module:
        skip = True
        return skip

    return skip


def setup(sphinx):
    sphinx.connect("autoapi-skip-member", skip_methods)



from typing import List, Tuple

import torch


class Embedding(torch.autograd.Function):
    def __init__(
        self, emb_name: str, emb_id: str, emb_dim: int, emb_size: int, initialization
    ):
        self.emb_name = emb_name
        self.emb_size = emb_size
        self.emb_dim = emb_dim
        self.emb_id = emb_id  # some emb_id share the feature space
        self.initialization = initialization

    def register(self):
        # TODO: implement register function for py_client
        # from persia.prelude import register_embedding
        ...


class VarLenEmbedding(Embedding):
    def __init__(
        self,
        emb_name: str,
        emb_id: str,
        emb_dim: int,
        emb_size: int,
        initialization,
        sample_fixed_size: int,
    ):
        super(VarLenEmbedding, self).__init__(
            emb_name, emb_id, emb_dim, emb_size, initialization
        )

    @staticmethod
    def forward(self):
        ...

    @staticmethod
    def backward(self):
        ...


class SumEmbedding(Embedding):
    def __init__(
        self,
        emb_name: str,
        emb_id: str,
        emb_dim: int,
        emb_size: int,
        initialization,
    ):
        super(SumEmbedding, self).__init__(
            emb_name, emb_id, emb_dim, emb_size, initialization
        )

    @staticmethod
    def forward(self, tensor):
        ...

    @staticmethod
    def backward(self, grad_output):
        ...


class EmbeddingFunction(torch.autograd.Function):
    def __init__(self, embedding_specs: List[Embedding]):
        self.embedding_specs = embedding_specs

    @staticmethod
    def forward(ctx, sparses_raw: List):
        ...

    @staticmethod
    def backward(self, grad_output):
        ...



import numpy as np

from tqdm import tqdm


def generate_loader(
    continuous_data,
    categorical_data,
    target,
    categorical_columns,
    batch_size: int = 128,
    skip_last_batch: bool = False,
):
    dataset_size = len(target)
    for start in range(0, dataset_size, batch_size):
        end = min(start + batch_size, dataset_size)
        if end == dataset_size and skip_last_batch:
            print("skip last batch...")
            continue
        continuous_batch_data = continuous_data[start:end, :]
        sparse_batch_data = []

        for idx, categorical_column in enumerate(categorical_columns):
            tmp_sparse_batch = []
            categorical_batch_data = categorical_data[start:end]
            for i in range(len(categorical_batch_data)):
                tmp_sparse_batch.append(categorical_batch_data[i][idx : idx + 1])
            sparse_batch_data.append((categorical_column, tmp_sparse_batch))

        target_batch = target[start:end]
        target_batch = target_batch.reshape(len(target_batch), -1)

        yield continuous_batch_data, sparse_batch_data, target_batch


def make_dataloader(
    data_filepath: str, batch_size: int = 128, skip_last_batch: bool = False
):
    with np.load(data_filepath) as data:
        target = data["target"]
        continuous_data = data["continuous_data"]
        categorical_data = data["categorical_data"]
        categorical_columns = data["categorical_columns"]

    dataset_size = len(target)
    loader_size = (dataset_size - 1) // batch_size + 1
    if skip_last_batch:
        loader_size = loader_size - 1

    return loader_size, generate_loader(
        continuous_data,
        categorical_data,
        target,
        categorical_columns,
        batch_size,
        skip_last_batch,
    )


if __name__ == "__main__":
    loader = make_dataloader("data_source/train.npz", 128, skip_last_batch=False)
    for (dense, sparse, target) in tqdm(loader, desc="generate_data"):
        ...



import os
import time
import yaml

from persia.error import FileNotFoundException


def setup_seed(seed: int):
    """Set the seed for most used common library to ensure the experiment reproducible

    Arguments:
        seed (int): Seed of common library.
    """
    import numpy as np
    import torch
    import random

    np.random.seed(seed)

    random.seed(seed)

    torch.random.manual_seed(seed)
    if getattr(torch, "use_deterministic_algorithms", None):
        torch.use_deterministic_algorithms(True)
    else:
        torch.backends.cudnn.deterministic = True


def load_yaml(filepath: str) -> dict:
    """load the yaml config by provided filepath

    Arguments:
        filepath (str): yaml config path
    """
    if os.path.exists(filepath):
        raise FileNotFoundException(f"filepath {filepath} not found!")

    with open(filepath, "r") as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def block(interval: int = 120):
    """block the process by sleep function

    Arguments:
        interval (int, optinal): sleep interval
    """
    while True:
        time.sleep(interval)



import torch
import torch.nn as nn


class DNN(nn.Module):
    def __init__(
        self, dense_mlp_output_size: int = 16, sparse_mlp_output_size: int = 128
    ):
        super(DNN, self).__init__()

        self.dense_mlp = torch.nn.Linear(5, dense_mlp_output_size)
        self.dense_bn = nn.BatchNorm1d(dense_mlp_output_size)

        self.sparse_mlp = torch.nn.Linear(64, sparse_mlp_output_size)
        self.sparse_bn = nn.BatchNorm1d(sparse_mlp_output_size)

        self.ln1 = nn.Linear(dense_mlp_output_size + sparse_mlp_output_size, 256)
        self.ln2 = nn.Linear(256, 128)
        self.ln3 = nn.Linear(128, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, dense_x, sparses):
        sparse_concat = torch.cat(sparses, dim=1)
        sparse = self.sparse_mlp(sparse_concat.float())
        sparse = self.sparse_bn(sparse)

        dense_x = self.dense_mlp(dense_x)
        dense_x = self.dense_bn(dense_x)
        x = torch.cat([sparse, dense_x], dim=1)
        x = self.ln1(x)
        x = self.ln2(x)
        x = self.ln3(x)

        return self.sigmoid(x)


if __name__ == "__main__":
    model = DNN()
    batch_size = 64
    dense = torch.ones(batch_size, 5)
    sparses = [torch.ones(batch_size, 8) for _ in range(8)]
    output = model(dense, sparses)
    print(output)



from typing import List

from persia.prelude import (
    PyPersiaRpcClient,
    PyPersiaBatchData,
    PyPersiaReplicaInfo,
    PyPersiaBatchFlowNatsStubPublisher,
    PyOptimizerBase,
)
from persia.error import PersiaRuntimeException
from persia.env import get_world_size

_backend = None


class Backend:
    r"""Backend class is the python wrapper of PersiaRpcClient and NatsPublisher that provide the ability to
    invoke rpc function to communicate with the trainer and middleware componenet.

    Arguments:
        threadpool_worker_size (int): Rpc client threadpool size.
        replica_info (PyPersiaReplicaInfo): Replica info of current process.
    """

    def __init__(self, threadpool_worker_size: int, replica_info: PyPersiaReplicaInfo):
        self.rpc_client = PyPersiaRpcClient(threadpool_worker_size)

        world_size = get_world_size()
        self.nats_publisher = PyPersiaBatchFlowNatsStubPublisher(
            replica_info, world_size if world_size != -1 else None
        )
        self.nats_publisher.wait_servers_ready()

    def send_data(self, data: PyPersiaBatchData, blocking: bool = True):
        """Send PersiaBatchData from data compose to trainer and middleware side.

        Arguments:
            data (PyPersiaBatchData): PersiaBatchData that haven't been process.
            blocking (bool, optional): Wait util the data send successfully.
        """
        self.nats_publisher.send_sparse_to_middleware(data, blocking)
        self.nats_publisher.send_dense_to_trainer(data, blocking)

    def get_embedding_size(self) -> List[int]:
        """Get current embedding server embedding size"""

        return self.rpc_client.get_embedding_size()

    def set_configuration(
        self,
        initialize_lower: float,
        initialize_upper: float,
        admit_probability: float,
        enable_weight_bound: bool,
        weight_bound: float,
    ):
        """Set the configuration of embedding initialization and embedding update rule to embedding server.

        Arguments:
            initialize_lower (float): embedding uniform initialization lower bound args.
            initialize_upper (float): embedding uniform initialization upper bound args.
            admit_probability (float): probability of embedding generation. value in [0, 1]. Generate a random value(range in [0, 1]) for each sparse embedding,
                    generate the new embedding once the value is small than the admit_probability. Always generate the new embedding when the admit_probability set
                    to 1.
            enable_weight_bound (bool): Enable weight bound or not.
            weight_bound (float): Restrict each element value of an embedding in [-weight_bound, weight_bound].
        """
        self.nats_publisher.configure_sharded_servers(
            initialize_lower,
            initialize_upper,
            admit_probability,
            enable_weight_bound,
            weight_bound,
        )

    def register_optimizer(self, optimizer: PyOptimizerBase):
        """Register the sparse optimizer to embedding server.

        Arguments:
            optimizer (PyOptimizerBase): optimizer python wrapper.
        """
        self.nats_publisher.register_optimizer(optimizer)

    def dump_embedding(self, dst_dir: str, blocking: bool = False):
        """Dump the sparse embedding to destination directory.This method provide not blocking
        mode, invoke the TrainCtx.wait_for_dump_embedding once set blocking to False.

        Arguments:
            dst_dir (str): Destination directory.
            blocking (bool, optional): Dump embedding util finished if `blocking` set to `True`.
        """
        self.rpc_client.dump_embedding(dst_dir)
        if blocking:
            self.rpc_client.wait_for_dump_embedding()

    def load_embedding(self, src_dir: str, blocking: bool = True):
        """Load the sparse embedding from source directory by invoke rpc client.This method provide
        not blocking mode,invoke the TrainCtx.wait_for_load_embedding once set the blocking to False.

        Arguments:
            src_dir (str): Source directory.
            blocking (bool, optional): Load embedding util finished if `blocking` set to `True`.
        """
        self.rpc_client.load_embedding(src_dir)
        if blocking:
            self.rpc_client.wait_for_load_embedding()

    def wait_for_dump_embedding(self):
        """Wait for dump the sparse embedding."""
        self.rpc_client.wait_for_dump_embedding()

    def wait_for_load_embedding(self):
        """Wait for load the sparse embedding."""
        self.rpc_client.wait_for_load_embedding()


def init_backend(
    threadpool_worker_size: int,
    replica_info: PyPersiaReplicaInfo,
) -> Backend:
    """Initialized the singleton Backend instance.

    Arguments:
        threadpool_worker_size (int, optional): Rpc client threadpool size.
        replica_info (PyPersiaReplicaInfo, optional): Replica info of current process.
    """
    global _backend
    if not _backend:
        _backend = Backend(threadpool_worker_size, replica_info)
    return _backend


def get_backend() -> Backend:
    """Get singleton Backend instance, raise PersiaRuntimeException once uninitialized the backend."""
    if not _backend:
        raise PersiaRuntimeException("init persia backend first")
    return _backend



from abc import abstractmethod, ABC
from typing import Tuple

from persia.prelude import PyOptimizerBase
from persia.backend import get_backend


class Optimizer(ABC):
    r"""Base optimizer to configurate the sparse embedding update behavior."""

    def __init__(self):
        self.optimizer_base = PyOptimizerBase()

    @abstractmethod
    def apply(self):
        r"""Abstraction method for optimizer that support register different type of optimizer."""
        ...

    def register_optimizer(self):
        """Register sparse optimizer to embedding server."""
        get_backend().register_optimizer(self.optimizer_base)


class SGD(Optimizer):
    r"""A wrapper to config the embedding-server SGD optimizer."""

    def __init__(self, lr: float, momentum: float = 0.0, weight_decay: float = 0.0):
        """
        Arguments:
            params(float): learning rate.
            momentum(float, optional): momentum factor.
            weight_decay(float, optional): parameters L2 penalty factor.
        """
        super(SGD, self).__init__()
        self.lr = lr
        self.momentum = momentum
        self.weight_decay = weight_decay

    def apply(self):
        """Initialize optimizer and register to embedding server."""
        self.optimizer_base.init_sgd(self.lr, self.weight_decay)
        self.register_optimizer()


class Adam(Optimizer):
    r"""A wrapper to config the embedding-server Adam optimizer."""

    def __init__(
        self,
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        weight_decay: float = 0,
        eps: float = 1e-8,
    ):
        """
        Arguments:
            lr(float): learning rate.
            betas(tuple[float,float], optional): caculate the running averages of gradient and its square.
            weight_decay(float, optional): parameters L2 penalty factor.
            eps(float, optional): epsilon to avoid div zero.
        """
        super(Adam, self).__init__()
        self.lr = lr
        self.betas = betas
        self.weight_decay = weight_decay
        self.eps = eps

    def apply(self):
        """Initialize Adam optimizer and register to embedding server."""
        self.optimizer_base.init_adam(self.lr, self.betas, self.eps)
        self.register_optimizer()


class Adagrad(Optimizer):
    r"""A wrapper to config the embedding-server Adagrad optimizer."""

    def __init__(
        self,
        lr: float = 1e-2,
        initial_accumulator_value: float = 1e-2,
        weight_decay: float = 0,
        g_square_momentum: float = 1,
        eps: float = 1e-10,
    ):
        """
        Arguments:
            lr (float): learning rate.
            initial_accumulator_value (float, optional): initialization accumulator value for adagrad optimizer.
            weight_decay (float, optional): parameters L2 penalty factor.
            g_square_momentum (float, optional): factor of accumulator incremental.
            eps(float, optional): epsilon term to avoid divide zero.

        """
        super(Adagrad, self).__init__()
        self.lr = lr
        self.weight_decay = weight_decay
        self.initial_accumulator_value = initial_accumulator_value
        self.g_square_momentum = g_square_momentum
        self.eps = eps

    def apply(self):
        """Initialize Adagrad optimizer and register to embedding server."""
        self.optimizer_base.init_adagrad(
            self.lr,
            self.weight_decay,
            self.initial_accumulator_value,
            self.g_square_momentum,
            self.eps,
        )
        self.register_optimizer()



import os

from argparse import ArgumentParser

import pandas as pd
import numpy as np

from sklearn.preprocessing import OrdinalEncoder


def process(df_dataset: pd.core.frame.DataFrame, filepath: str):

    # categoricasl data encoded to int value
    for col in CATEGORICAL_COLUMNS:
        encoder = OrdinalEncoder()
        df_dataset[col] = encoder.fit_transform(df_dataset[[col]]).astype(np.uint64)

    categorical_data = np.vstack([df_dataset[k].values for k in CATEGORICAL_COLUMNS]).T
    continuous_data = np.vstack([df_dataset[k].values for k in CONTINUOUS_COLUMNS]).T

    df_dataset["target"] = df_dataset["income_bracket"].apply(lambda x: ">50K" in x)
    target = df_dataset["target"].to_numpy()

    np.savez_compressed(
        filepath,
        target=target.astype(np.float32),
        continuous_data=continuous_data.astype(np.float32).copy(),
        categorical_data=categorical_data,
        categorical_columns=CATEGORICAL_COLUMNS,
    )


if __name__ == "__main__":
    COLUMNS = [
        "age",
        "workclass",
        "fnlwgt",
        "education",
        "education_num",
        "marital_status",
        "occupation",
        "relationship",
        "race",
        "gender",
        "capital_gain",
        "capital_loss",
        "hours_per_week",
        "native_country",
        "income_bracket",
    ]

    CATEGORICAL_COLUMNS = [
        "workclass",
        "education",
        "marital_status",
        "occupation",
        "relationship",
        "race",
        "gender",
        "native_country",
    ]

    CONTINUOUS_COLUMNS = [
        "age",
        "education_num",
        "capital_gain",
        "capital_loss",
        "hours_per_week",
    ]

    parser = ArgumentParser()
    parser.add_argument("--train-dataset", default="data_source/train.csv")
    parser.add_argument("--test-dataset", default="data_source/test.csv")
    parser.add_argument("--output_path", default="data_source")
    args = parser.parse_args()

    output_path = args.output_path

    if args.train_dataset:
        df_train = pd.read_csv(args.train_dataset, names=COLUMNS, skipinitialspace=True)
        process(df_train, os.path.join(output_path, "train.npz"))

    if args.test_dataset:
        df_test = pd.read_csv(
            args.test_dataset, names=COLUMNS, skipinitialspace=True, skiprows=1
        )
        process(df_test, os.path.join(output_path, "test.npz"))



import sys

from types import ModuleType

# pytype: disable=import-error
import persia_core

# pytype: enable=import-error


def register_submodule(module: ModuleType, root_module_path: str):
    """register the persia core module to sys module path.

    Arguments:
        module (ModuleType): Root module.
        root_module_path (str): Root module path.
    """
    for attr in dir(module):
        if attr.startswith("__"):
            continue
        obj = getattr(module, attr)
        if isinstance(obj, ModuleType):
            submodule_name = attr
            full_path = f"{root_module_path}.{submodule_name}"
            sys.modules[full_path] = obj
            register_submodule(obj, full_path)


register_submodule(
    persia_core,
    persia_core.__name__,
)

# pytype: disable=import-error
from persia_core import (
    is_cuda_feature_available,
    init_persia_embedding_staleness_semaphore,
)

from persia_core import (
    PyPersiaRpcClient,
)
from persia_core.optim import PyOptimizerBase
from persia_core.data import (
    PyPersiaBatchData,
)
from persia_core.utils import (
    PyPersiaMessageQueueServer,
    PyPersiaMessageQueueClient,
    PyPersiaBatchDataChannel,
    PyPersiaBatchDataSender,
    PyPersiaBatchDataReceiver,
    PyPersiaReplicaInfo,
)
from persia_core.nats import PyPersiaBatchFlowNatsStubPublisher, init_responder

if is_cuda_feature_available():
    from persia_core.backward import PyBackward
    from persia_core.forward import (
        PyForward,
        forward_directly_from_data,
        forward_directly_from_bytes,
    )

# pytype: enable=import-error



import os

import torch
import numpy as np

from tqdm import tqdm
from sklearn import metrics

from persia.ctx import TrainCtx, eval_ctx
from persia.sparse.optim import Adagrad
from persia.env import get_rank, get_local_rank, get_world_size
from persia.logger import get_default_logger
from persia.utils import setup_seed
from persia.data import Dataloder, PersiaDataset, StreamingDataset
from persia.prelude import PyPersiaBatchData, PyPersiaBatchDataSender

from model import DNN
from data_generator import make_dataloader


logger = get_default_logger("trainer")

device_id = get_local_rank()

setup_seed(3)


class TestDataset(PersiaDataset):
    def __init__(self, test_dir: str, batch_size: int = 128):
        super(TestDataset, self).__init__(buffer_size=10)
        size, loader = make_dataloader(test_dir, batch_size)
        self.loader = loader
        self.loader_size = size

        logger.info(f"test dataset size is {size}")

    def fetch_data(self, persia_sender_channel: PyPersiaBatchDataSender):
        logger.info("test loader start to generate data...")
        for idx, (dense, batch_sparse_ids, target) in enumerate(
            tqdm(self.loader, desc="gen batch data")
        ):
            batch_data = PyPersiaBatchData()
            batch_data.add_dense([dense])
            batch_data.add_sparse(batch_sparse_ids)
            batch_data.add_target(target)
            persia_sender_channel.send(batch_data)

    def __len__(self):
        return self.loader_size


def test(model: torch.nn.Module, data_laoder: Dataloder):
    logger.info("start to test...")
    model.eval()

    with eval_ctx() as ctx:
        accuracies, losses = [], []
        all_pred, all_target = [], []
        for (batch_idx, batch_data) in enumerate(tqdm(data_laoder, desc="test...")):
            dense, sparse, target = ctx.prepare_features(batch_data)
            output = model(dense, sparse)
            loss = loss_fn(output, target)
            all_pred.append(output.cpu().detach().numpy())
            all_target.append(target.cpu().detach().numpy())
            accuracy = (torch.round(output) == target).sum() / target.shape[0]
            accuracies.append(accuracy)
            losses.append(float(loss))

        all_pred, all_target = np.concatenate(all_pred), np.concatenate(all_target)

        fpr, tpr, th = metrics.roc_curve(all_target, all_pred)
        test_auc = metrics.auc(fpr, tpr)

        test_accuracy = torch.mean(torch.tensor(accuracies))
        test_loss = torch.mean(torch.tensor(losses))
        logger.info(
            f"test auc is {test_auc} accuracy is {test_accuracy}, loss is {test_loss}"
        )

    model.train()

    return test_auc, test_accuracy


if __name__ == "__main__":
    model = DNN()
    logger.info("init Simple DNN model...")
    rank, device_id, world_size = get_rank(), get_local_rank(), get_world_size()

    torch.cuda.set_device(device_id)
    model.cuda(device_id)

    dense_optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
    sparse_optimizer = Adagrad(lr=1e-2)
    loss_fn = torch.nn.BCELoss(reduction="mean")
    logger.info("finish genreate dense ctx")

    test_dir = os.path.join("/data/test.npz")
    test_dataset = TestDataset(test_dir, batch_size=128)
    test_interval = 254
    buffer_size = 10

    with TrainCtx(
        sparse_optimizer=sparse_optimizer,
        dense_optimizer=dense_optimizer,
        device_id=device_id,
    ) as ctx:
        train_dataloader = Dataloder(StreamingDataset(buffer_size))
        test_loader = Dataloder(test_dataset, is_training=False)

        for (batch_idx, data) in enumerate(train_dataloader):
            dense, sparse, target = ctx.prepare_features(data)
            output = model(dense, sparse)
            loss = loss_fn(output, target)
            scaled_loss = ctx.backward(loss)
            accuracy = (torch.round(output) == target).sum() / target.shape[0]
            logger.info(
                f"current idx: {batch_idx} loss: {loss} scaled_loss: {scaled_loss} accuracy: {accuracy}"
            )

            if batch_idx % test_interval == 0 and batch_idx != 0:
                test_auc, test_acc = test(model, test_loader)
                assert (
                    test_auc > 0.8
                ), f"test_auc error, expect greater than 0.8 but got {test_auc}"
                break



import os

from enum import Enum
from queue import Queue
from typing import List, Tuple, Optional, NewType

import torch

import persia.env as env

from persia.logger import get_default_logger
from persia.sparse.optim import Optimizer
from persia.backend import init_backend
from persia.prelude import (
    PyPersiaReplicaInfo,
    init_persia_embedding_staleness_semaphore,
)

_CURRENT_CXT = None

_logger = get_default_logger()

PythonTrainBatch = NewType("PythonTrainBatch", object)


def _check_finite(tensors: List[torch.Tensor]) -> bool:
    """Check all tensors in the input list contain only finite elements.

    Arguments:
        tensors (List[torch.Tensor]): tensor list that need to check finite or not.

    Returns: Whether all the tensors is finite or is None.
    """
    return all([torch.isfinite(t).all() if t is not None else True for t in tensors])


class PreprocessMode(Enum):
    r"""Different preprocess mode will effect the ``EmbeddingCtx.prepare_features`` return result. ``PreprocessMode.TRAIN`` will return
    the torch tensor that the ``requires_grad`` attribute is set to ``True``. ``EmbeddingCtx.EVAL`` will return the torch tensor
    that the ``requires_grad`` attribute is set to ``False``. ``EmbeddingCtx.INFERENCE``  almost behave like ``EmbeddingCtx.EVAL``, the only difference
    is that ``PreprocessMode.INFERENCE`` allows ""EmbeddingCtx`` to process the ``PythonTrainBatch`` without target tensor.
    """
    TRAIN = 1
    EVAL = 2
    INFERENCE = 3


class BaseCtx:
    r"""It provide the communicate ability for data generator component to send the PersiaBatchData
    to the trainer and embedding middleware.

    Examples::
        >>> from persia.prelude import PyPersiaBatchData
        >>> loader = make_simple_loader()
        >>> with BaseCtx() as ctx:
        >>>     for (dense, batch_sparse_ids, target) in loader:
        >>>         batch_data = PyPersiaBatchData()
        >>>         batch_data.add_dense([dense])
        >>>         batch_data.add_sparse(batch_sparse_ids)
        >>>         batch_data.add_target(target)
        >>>         ctx.backend.send_data(batch_data)
    """

    def __init__(
        self,
        threadpool_worker_size: int = 10,
    ):
        """
        Arguments:
            threadpool_worker_size (int): Rpc threadpool worker size.
        """
        world_size = env.get_world_size()

        if world_size == -1:
            replica_size = env.get_replica_size()
            replica_index = env.get_replica_index()
            self.replica_info = PyPersiaReplicaInfo(replica_size, replica_index)
            _logger.info(
                f"init datacompose backend replica_size: {replica_size} replica_index: {replica_index}"
            )
        else:
            rank_id = env.get_rank()
            self.replica_info = PyPersiaReplicaInfo(world_size, rank_id)
            _logger.info(
                f"init trainer backend world size: {world_size} rank_id: {rank_id}"
            )

        self.backend = init_backend(threadpool_worker_size, self.replica_info)

    def _enter(self):
        """Hook when enter the context"""
        ...

    def _exit(self):
        """Hook when exit the context"""
        ...

    def __enter__(self):
        self._enter()

        global _CURRENT_CXT
        _CURRENT_CXT = self

        return self

    def __exit__(self, exc_type, value, trace):
        self._exit()

        global _CURRENT_CXT
        _CURRENT_CXT = None

        if exc_type:
            import traceback

            _logger.error("\n" + traceback.format_exc())


class EmbeddingCtx(BaseCtx):
    r"""EmbeddingCtx provide the embedding relative function compare to BaseCtx.It can run the offline test or online inference
    according to different preprocess_mode.The most simple way to get this context is use ``persia.ctx.eval_ctx()`` or
    ``persia.ctx.inference_ctx`` to get the ``EmbeddingCtx`` instance.

    Examples::
        >>> from persia.prelude import PyPersiaBatchData
        >>> model = get_dnn_model()
        >>> loader = make_dataloader()
        >>> with EmbeddingCtx(
        ...     PreprocessMode.EVAL
        ... ) as ctx:
        >>>     for (dense, batch_sparse_ids, target) in loader:
        >>>         batch_data = PyPersiaBatchData()
        >>>         batch_data.add_dense([dense])
        >>>         batch_data.add_sparse(batch_sparse_ids)
        >>>         batch_data.add_target(target)
        >>>         python_train_batch = forward_directly_from_data(batch_data)
        >>>         dense_tensor, sparse_tensors, target_tensor = ctx.prepare_features(python_train_batch)
        >>>         output = model(dense_tensor, sparse_tensors)
    """

    def __init__(
        self,
        preprocess_mode: PreprocessMode,
        emb_initialization: Tuple[float, float] = (-0.01, 0.01),
        admit_probability: float = 1.0,
        weight_bound: float = 10,
        checkpoint_dir: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """
        Arguments:
            preprocess_mode (PreprocessMode): Different preprocess mode effect the behave of ``prepare_features``.
            emb_initialization (Tuple[float, float], optional): Embedding uniform initialization arguments that corresponding to low and high.
            admit_probability (float, optional): The probability (0<=, <=1) of admitting a new embedding.
            weight_bound (float, optional): Restrict each element value of an embedding in [-weight_bound, weight_bound].
            checkpoint_dir(str, optional): Pretrained checkpoint directory, load the dense and sparse checkpoint in this dir when enter the context.
        """
        super(EmbeddingCtx, self).__init__(*args, **kwargs)
        self.preprocess_mode = preprocess_mode
        self.emb_initialization = emb_initialization
        self.admit_probability = admit_probability
        self.weight_bound = weight_bound
        self.checkpoint_dir = checkpoint_dir

        self.current_batch = None
        self.pretrained_loaded = False

    def _enter(self):
        self.backend.set_configuration(
            self.emb_initialization[0],
            self.emb_initialization[1],
            self.admit_probability,
            self.weight_bound > 0,
            self.weight_bound,
        )

        if not self.pretrained_loaded and self.checkpoint_dir is not None:
            self.load_embedding(self.checkpoint_dir)
            self.pretrained_loaded = True

    def prepare_features(
        self, batch: PythonTrainBatch
    ) -> Tuple[torch.Tensor, List[torch.Tensor], Optional[torch.Tensor]]:
        """Converted the dense, sparse and target raw data in``PythonTrainBatch`` to `torch.Tensor``.

        Arguments:
            batch (PythonTrainBatch): Training data provided by PersiaML upstream including
                dense, target, sparse data and meta info.

        Returns:
            the tuple of dense data, list of sparse data and target data.
        """
        import persia_torch_ext as pte  # pytype: disable=import-error

        if self.preprocess_mode == PreprocessMode.INFERENCE:
            batch.target_tensor = None
        else:
            # pytype: disable=attribute-error
            batch.target = batch.consume_all_targets()
            # pytype: enable=attribute-error
            assert len(batch.target) == 1
            batch.target = batch.target[0]

            batch.target_tensor = pte.ptr_to_tensor_f32(
                batch.target.data_ptr(), batch.target.shape(), False
            )

        is_training = self.preprocess_mode == PreprocessMode.TRAIN  # cache property

        # pytype: disable=attribute-error
        batch.dense = batch.consume_all_dense_features()
        # pytype: enable=attribute-error
        batch.dense = batch.dense[0]
        batch.dense_tensor = pte.ptr_to_tensor_f32(
            batch.dense.data_ptr(), batch.dense.shape(), False
        )

        # pytype: disable=attribute-error
        batch.emb = batch.consume_all_sparse_features()
        # pytype: enable=attribute-error

        batch.emb_slot = []
        # sparse embedding processing
        emb_tensors, forward_tensors = [], []

        for emb in batch.emb:
            if emb.is_raw_embedding():
                # no duplicate id in raw_id_tensor
                (
                    raw_embedding,
                    index,
                    non_empty_index,
                    sample_id_num,
                ) = emb.get_raw_embedding()

                batch.emb_slot.append([raw_embedding, index, non_empty_index])

                distinct_id_tensor = pte.ptr_to_tensor_f16(
                    raw_embedding.data_ptr(), raw_embedding.shape(), False
                )
                index_tensor = pte.ptr_to_tensor_long(
                    index.data_ptr(),
                    index.shape(),
                )  # tensor shape (1, batch_size * sample_fixed_size)
                max_index = index_tensor.max()
                size_of_distinct_id_tensor = distinct_id_tensor.shape[0]
                torch.cuda.synchronize()

                assert (
                    max_index < size_of_distinct_id_tensor
                ), "raw embedding select index larger than tensor"
                non_empty_index_tensor = pte.ptr_to_tensor_long(
                    non_empty_index.data_ptr(), non_empty_index.shape()
                )  # tensor shape (-1), variable length

                batch_size = len(sample_id_num)
                dim = distinct_id_tensor.shape[-1]
                sample_fixed_size = index_tensor.shape[1] // batch_size
                index_select_raw_tensor = distinct_id_tensor.index_select(
                    0, index_tensor.view(-1)
                )
                index_select_raw_tensor.requires_grad = is_training

                raw_fixed_size_tensor = index_select_raw_tensor.view(
                    -1, sample_fixed_size, dim
                )
                mask = (
                    index_tensor.view(batch_size, sample_fixed_size, 1) != 0
                ).half()  # generate mask
                raw_fixed_size_tensor_with_mask = torch.cat(
                    [raw_fixed_size_tensor, mask], dim=2
                )
                emb_tensors.append(
                    (
                        raw_embedding.name(),
                        distinct_id_tensor,
                        index_tensor,
                        non_empty_index_tensor,
                        index_select_raw_tensor,
                    )
                )
                forward_tensors.append(raw_fixed_size_tensor_with_mask)
            else:
                emb = emb.get_sum_embedding()
                batch.emb_slot.append([emb])

                sum_tensor = pte.ptr_to_tensor_f16(
                    emb.data_ptr(), emb.shape(), is_training
                )
                forward_tensors.append(sum_tensor)
                emb_tensors.append((emb.name(), None, None, None, sum_tensor))

        batch.forward_tensors = forward_tensors
        batch.emb_tensors = emb_tensors
        self.current_batch = batch

        return batch.dense_tensor, batch.forward_tensors, batch.target_tensor

    def dump_checkpoint(
        self,
        dst_dir: str,
        model: Optional[torch.nn.Module] = None,
        optimizer: Optional[torch.optim.Optimizer] = None,
        default_filename: str = "dense.pt",
        blocking: bool = True,
    ):
        """Dump the dense and sparse checkpoint to destination directory.

        Arguments:
            dst_dir (str): Destination directory.
            model (torch.nn.Module, optional): Pytorch model instance.
            optimizer (torch.optim.Optimizer, optional): Pytorch optimizer instance.
            default_filename (str, optional): Dense checkpoint filename.
            blocking (bool, optional): Dump embedding checkpoint in blocking mode or not.
        """
        os.makedirs(dst_dir, exist_ok=True)

        cpk_dict = {}
        if model:
            cpk_dict["model"] = model.state_dict()

        if optimizer:
            cpk_dict["opt"] = optimizer.state_dict()

        if len(cpk_dict.keys()) > 0:
            dense_model_filepath = os.path.join(dst_dir, default_filename)
            torch.save(cpk_dict, dense_model_filepath)

        self.dump_embedding(dst_dir, blocking=blocking)

    def load_checkpoint(
        self,
        src_dir: str,
        model: Optional[torch.nn.Module] = None,
        optimizer: Optional[torch.optim.Optimizer] = None,
        map_location: Optional[str] = None,
        default_filename: str = "dense.pt",
        blocking: bool = True,
    ):
        """Load the dense and sparse checkpoint from source directory.

        Arguments:
            src_dir (str): Source directory.
            model (torch.nn.Module, optional): Pytorch model instance.
            optimizer (torch.optim.Optimizer, optional): Pytorch optimizer instance.
            map_location (str, optional): Load the dense checkpoint to specific device.
            default_filename (str, optional): Dense checkpoint filename.
            blocking (bool, optional): Dump embedding checkpoint in blocking mode or not.
        """
        if not os.path.exists(src_dir):
            _logger.warn(f"source directory: {src_dir} not exists")
            return

        dense_model_filepath = os.path.join(src_dir, default_filename)
        if os.path.exists(dense_model_filepath):
            cpk_dict = torch.load(dense_model_filepath, map_location)

            if model:
                model.load_state_dict(cpk_dict["model"], strict=True)

            if optimizer:
                optimizer.load_state_dict(cpk_dict["opt"])

        self.load_embedding(src_dir, blocking=blocking)

    def dump_embedding(self, dst_dir: str, blocking: bool = False):
        """Dump embeddings to ``dst_dir``. Use ``TrainCtx.wait_for_dump_embedding`` to wait until finished
        if ``blocking=False``.

        Arguments:
            dst_dir (str): Destination directory.
            blocking (bool, optional): Dump embedding in blocking mode or not.
        """
        self.backend.dump_embedding(dst_dir, blocking)

    def load_embedding(self, src_dir: str, blocking: bool = True):
        """Load embeddings from ``src_dir``. Use ``TrainCtx.wait_for_load_embedding`` to wait until finished
        if ``blocking=False``.

        Arguments:
            src_dir (str): Directory to load embeddings.
            blocking (bool, optional): Dump embedding in blocking mode or not.
        """
        self.backend.load_embedding(src_dir, blocking)

    def wait_for_dump_embedding(self):
        """Wait for the embedding dump process."""
        self.backend.wait_for_dump_embedding()

    def wait_for_load_embedding(self):
        """Wait for the embedding load process."""
        self.backend.wait_for_load_embedding()


class TrainCtx(EmbeddingCtx):
    r"""Subclass of ``EmbeddingCtx`` that provide the backward ability to update the sparse embedding.

    Example::
        >>> import torch
        >>> model = get_dnn_model()
        >>> sparse_optimizer = persia.sparse.optim.SGD(lr=1e-3)
        >>> dense_optimizer = torch.optim.SGD(lr=1e-3)
        >>> loss_fn = torch.nn.BCELoss(reduction="mean")
        >>> with TrainCtx(
        >>>     sparse_optimizer,
        >>>     dense_optimizer,
        >>>     mixed_precision=True
        >>> ) as ctx:
        >>>     for batch_data in dataloder:
        >>>         dense, sparse, target = ctx.prepare_features(data)
        >>>         output = model(dense, sparse)
        >>>         loss = loss_fn(output, target)
        >>>         scaled_loss = ctx.backward(loss)
    """

    def __init__(
        self,
        sparse_optimizer: Optimizer,
        dense_optimizer: torch.optim.Optimizer,
        device_id: int = 0,
        grad_scalar_update_factor: float = 4,
        embedding_staleness: int = -1,
        backward_buffer_size: int = 10,
        backward_workers_size: int = 8,
        grad_update_buffer_size: int = 60,
        *args,
        **kwargs,
    ):
        """
        Arguments:
            sparse_optimizer (persia.sparse.optim.Optimizer): Optimizer for the embeddings.
            dense_optimizer (torch.optim.Optimizer): Optimizer for dense parameters.
            device_id (int, optional): The CUDA device to use for this process.
            grad_scalar_update_factor (float, optional): Update factor of ``Gradscalar`` to ensure loss scale finitely if set ``mixed_precision=True``.
            embedding_staleness (int, optional): Max number of batched staleness embedding each rank. A staleness embedding means it prefetched from embedding server before gradient updated.
            backward_buffer_size (int, optional): Max number of not updated gradients queued.
            backward_workers_size (int, optional): Number of workers sending embedding gradients in parallel.
            grad_tensor_cache_size(int, optional): Number of reference cache , hold the gradient tensor reference to avoid
                meet dangle data in gradient backward phase.
        """
        super(TrainCtx, self).__init__(PreprocessMode.TRAIN, *args, **kwargs)

        assert (
            sparse_optimizer is not None
        ), "Sparse_optimizer should not be none in train context"
        assert (
            0 <= device_id < torch.cuda.device_count()
        ), f"device_id: {device_id} invalid!"
        assert grad_scalar_update_factor > 0, "grad scalar should greater than zero"

        torch.cuda.set_device(device_id)

        self.device_id = device_id

        self.update_times = 0
        self.grad_scalar_update_factor = grad_scalar_update_factor
        self.grad_scaler = torch.cuda.amp.GradScaler()
        self.sparse_optimizer = sparse_optimizer
        self.dense_optimizer = dense_optimizer
        self.backward_workers_size = backward_workers_size
        self.embedding_staleness = (
            embedding_staleness if embedding_staleness is not None else -1
        )

        from persia.prelude import PyBackward

        # dynamic import the PyForward due to conditional compilation
        self.grad_queue = Queue(grad_update_buffer_size)
        self.backward_engine = PyBackward(backward_buffer_size)

    def _enter(self):
        super()._enter()

        if self.embedding_staleness > 0:
            init_persia_embedding_staleness_semaphore(self.embedding_staleness)
        self.sparse_optimizer.apply()
        self.backward_engine.launch(self.device_id, self.backward_workers_size)

    def backward(
        self, loss: torch.Tensor, embedding_gradient_check_frequency: int = 20
    ) -> torch.Tensor:
        """Compute the gradient of current dense and sparse tensors.

        The backward support mixed precision training.According to current embedding gradient is finite or not, ``GradScalar``
        can update the scale automatic to restrict to parameters in finite range.

        Arguments:
            loss (torch.Tensor): Loss of current batch.
            embedding_gradient_check_frequency (int, optional): The frequency to check gradient finite or not for current embedding.
        """

        loss = self.grad_scaler.scale(loss)
        scale = self.grad_scaler.get_scale()

        loss.backward()

        finite = self._on_backward(scale, embedding_gradient_check_frequency)

        self.grad_scaler.step(self.dense_optimizer)

        if finite:
            self.grad_scaler.update()
        else:
            self.grad_scaler.update(scale / self.grad_scalar_update_factor)

        self.dense_optimizer.zero_grad()

        return loss

    def _on_backward(self, loss_scale: float, embedding_gradient_check_frequency: int):
        """Update the embeddings gradients

        Arguments:
            loss_scale (float): The loss that scaled by GradScalar.
            embedding_gradient_check_frequency (int): The frequency to check gradient finite or not for current embedding.
        """
        if self.grad_queue.full():
            self.grad_queue.get()

        finite = True
        if self.update_times % embedding_gradient_check_frequency == 0:
            finite = _check_finite(
                [emb[-1].grad for emb in self.current_batch.emb_tensors]
            )
            self.update_times += 1

        grad_slots, empty_grads = [], []
        gradient_batch = self.current_batch.create_gradient_batch()

        for (
            emb_name,
            distinct_id_tensor,
            index,
            non_zero_index,
            emb_tensor,
        ) in self.current_batch.emb_tensors:
            if emb_tensor.grad is None:
                gradient_batch.add_skipped_gradient(emb_name)
                empty_grads.append(emb_name)
            else:
                if distinct_id_tensor is not None:
                    if distinct_id_tensor.shape[0] > 1:
                        grad = torch.zeros_like(distinct_id_tensor, dtype=torch.float32)
                        non_zero_grad = emb_tensor.grad.index_select(
                            0, non_zero_index.view(-1)
                        ).float()
                        non_zero_index = index.view(-1)[non_zero_index.view(-1)]
                        grad.index_add_(0, non_zero_index, non_zero_grad)
                        grad = grad[1:, :]
                        is_f16_gradient = False
                    else:
                        grad = None
                else:
                    grad = emb_tensor.grad  # type: torch.Tensor
                    is_f16_gradient = True

                if grad is not None:
                    grad_slots.append(grad)
                    gradient_batch.add_gradient(
                        emb_name,
                        grad.data_ptr(),
                        grad.shape,
                        is_f16_gradient,
                        loss_scale,
                    )

        torch.cuda.synchronize()
        self.backward_engine.update_sparse_gradient_batched(gradient_batch)
        self.grad_queue.put(grad_slots)

        if len(empty_grads) > 0:
            _logger.warning(
                f"Current batch exists empty gradient tensors, num: {len(empty_grads)}, {empty_grads}"
            )
        return finite


def cnt_ctx() -> Optional[BaseCtx]:
    """Get the BaseCtx recently entered."""
    return _CURRENT_CXT


def eval_ctx(*args, **kwargs) -> EmbeddingCtx:
    """Get the ``EmbeddingCtx`` with the ``PreprocessMode.EVAL`` mode."""
    return EmbeddingCtx(PreprocessMode.EVAL, *args, **kwargs)


def inference_ctx(*args, **kwargs) -> EmbeddingCtx:
    """Get the ``EmbeddingCtx`` with the ``PreprocessMode.INFERENCE`` mode."""
    return EmbeddingCtx(PreprocessMode.INFERENCE, *args, **kwargs)



import os

from typing import List


def get_middleware_services() -> List[str]:
    """return the current middleware service list by environment variable"""
    return (
        [os.environ["MIDDLEWARE_SERVICE"]]
        if os.environ.get("MIDDLEWARE_SERVICE", None)
        else ["middleware:8000"]
    )


def get_trainer_services() -> List[str]:
    """return the current trainer service list by environment variable"""
    return (
        [os.environ["TRAINER_SERVICE"]]
        if os.environ.get("TRAINER_SERVICE", None)
        else ["trainer:8000"]
    )



import logging

from typing import Optional

from colorlog import ColoredFormatter


class levelFilter(logging.Filter):
    r"""log level filter to ensure log fileter

    Arguments:
        level (int): filter log level, remain the log greater than the setting level
    """

    def __init__(self, level: int):
        self.level = level

    def filter(self, record: logging.LogRecord) -> bool:
        """filter the record that level greater than the setting log level

        Arguments:
            record (logging.LogRecord): callback function input record items
        """
        return record.levelno > self.level


STREAM_LOG_FORMAT = "%(log_color)s%(asctime)s %(levelname)-8s%(reset)s %(blue)s[%(filename)s:%(lineno)d]%(reset)s %(log_color)s%(message)s"
FILE_LOG_FORMAT = "%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s"
DEFAULT_LOGGER_NAME = "log"
_default_logger = None

LOG_COLOR = {
    "DEBUG": "cyan",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "red,bg_white",
}

COLOR_FORMATTER = ColoredFormatter(
    STREAM_LOG_FORMAT,
    datefmt=None,
    reset=True,
    log_colors=LOG_COLOR,
    secondary_log_colors={},
    style="%",
)
FORMATTER = logging.Formatter(
    FILE_LOG_FORMAT,
    datefmt=None,
    style="%",
)


def setLogger(
    name: str,
    log_level: int = logging.DEBUG,
    log_filename: str = "train.log",
    enable_file_logger: bool = False,
    err_redirect_filepath: str = "error.log",
    enable_err_redirect: bool = False,
    err_redirect_level: int = logging.INFO,
) -> logging.Logger:
    r"""this function make logger configuration simplify, provide the
    log_level and log filename.It also make error log redirect available

    Arguments:
        name (str): logger name
        log_filename (str): log filename
        enable_file_logger (bool): whether enable save log into file
        err_redirect_filepath (str): err log redirect filepath
        enable_err_redirect (bool): whether enable err log redirect
        err_redirect_level (int): error redirect log level
    """
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    handler.setFormatter(COLOR_FORMATTER)

    logger.addHandler(handler)
    logger.setLevel(log_level)

    if enable_file_logger:
        file_normal_handler = logging.FileHandler(log_filename, mode="a")
        file_normal_handler.setFormatter(FORMATTER)
        logger.addHandler(file_normal_handler)

    if enable_err_redirect:
        file_error_handler = logging.FileHandler(err_redirect_filepath, mode="a")
        file_error_handler.setFormatter(FORMATTER)
        file_error_handler.addFilter(levelFilter(err_redirect_level))
        logger.addHandler(file_error_handler)
    return logger


def get_logger(name: str) -> logging.Logger:
    r"""get logger by name

    Arguments:
        name (str): logger name
    """
    return logging.getLogger(name)


def _set_default_logger(name: str, **kwargs) -> logging.Logger:
    r"""set default logger

    Arguments:
        name (str): default logger name

    logging.Logger
    """
    global _default_logger
    if not _default_logger:
        _default_logger = setLogger(name, **kwargs)
    return _default_logger


def get_default_logger(name: Optional[str] = None, **kwargs) -> logging.Logger:
    r"""get default logger or init the default by given name

    Arguments:
        name (str, optional): logger name
    """
    if _default_logger is None:
        _set_default_logger(name or DEFAULT_LOGGER_NAME, **kwargs)
    return _default_logger



from threading import Thread

import torch
from torch.utils.data.dataset import IterableDataset as TorchIterableDataset

from persia.ctx import cnt_ctx
from persia.logger import get_default_logger
from persia.prelude import (
    PyPersiaBatchDataChannel,
    PyPersiaReplicaInfo,
    PyPersiaBatchDataReceiver,
    PyPersiaBatchDataSender,
    init_responder,
)

_logger = get_default_logger()


class IterableDataset(
    TorchIterableDataset
):  # TODO: cannot understand the whole doc string
    r"""IterableChannelBase wrap the PyPersiaBatchDataChannel that provide the channel sender and
    receiver.

    Arguments:
        buffer_size (int): PyPersiaBatchDataChannel buffer size
    """

    def __init__(self, buffer_size: int):
        self.persia_batch_channel = PyPersiaBatchDataChannel(buffer_size)

    @property
    def receiver(self) -> PyPersiaBatchDataReceiver:
        """Get PyPersiaBatchDataReceiver python wrapper"""
        return self.persia_batch_channel.get_receiver()

    @property
    def sender(self) -> PyPersiaBatchDataSender:
        """Get PyPersiaBatchDataSender python wrapper"""
        return self.persia_batch_channel.get_sender()


class StreamingDataset(IterableDataset):
    r"""NatsStreamingChannel receive data from nats publisher

    Arguments:
        buffer_size (int): PyPersiaBatchDataChannel buffer size
    """

    def __init__(
        self,
        buffer_size: int,
    ):
        super(StreamingDataset, self).__init__(buffer_size)
        self.initialized = False

    def __iter__(self):
        if not self.initialized:
            current_ctx = cnt_ctx()
            assert current_ctx is not None, "Current conext is None!"
            replica_info = current_ctx.replica_info
            init_responder(replica_info, self.sender)

            _logger.info("initialize the responder")
            self.initialized = True

        while True:
            yield None


class PersiaDataset(IterableDataset):
    r"""Persia data channel that provide asynchronous data handler feature to improve the performance of data preprocess.
    Not support synchronous data handler temporary.

    Arguments:
        buffer_size (int): PyPersiaBatchDataChannel buffer size
        async_iterator (bool, optional): launch the thread to generate the data asynchronous
    """

    def __init__(
        self,
        buffer_size: int,
        async_iterator: bool = True,
    ):
        super(PersiaDataset, self).__init__(
            buffer_size,
        )
        self.async_iterator = async_iterator

    def fetch_data(self, sender: PyPersiaBatchDataSender):
        """Callback function to put the data into PyPersiaBatchDataSender

        Arguments:
            sender (PyPersiaBatchDataSender): PersiaBatchData sender channel to send the generate data
                to the PersiaBatchData receive channel
        """
        raise NotImplementedError("implement this function to fetch data")

    def __iter__(self):
        # TODO: provide process worker to handler the data with high performance calculating
        # class WorkerType(Enum):
        #     THREAD = 1
        #     PROCESS = 2

        if self.async_iterator:
            handler = Thread(target=self.fetch_data, args=(self.sender,), daemon=True)
            handler.start()

        for _val in range(len(self)):
            yield _val

        if self.async_iterator:
            handler.join()


class Dataloder(object):
    r"""Dataloder provide the interface to fetch the PythonBatchData from PyForward
    wrapper.

    Arguments:
        dataset (IterableChannelBase): dataset for Dataloder to retrive replica info and sender channel
        forward_buffer_size: (int, optional): gpu forward channel buffer size, this args effect the gpu memory cost
        is_training (bool, optional): whether current forward status is training or not
        timeout (int, optional): timeout for PyForward to fetch data, millisecond unit
        num_workers (int, optional): spawn thread worker number for  PyForward to lookup embedding and PythonBatchData prefetch
        reproducible (bool, optional): iterate the data in fixed order, make the dataflow deterministic
    """

    def __init__(
        self,
        dataset: IterableDataset,
        forward_buffer_size: int = 10,
        is_training: bool = True,
        timeout: int = 1000 * 60 * 10,
        num_workers: int = 10,
        replica_info: PyPersiaReplicaInfo = None,
        reproducible: bool = False,
    ):
        # dynamic import the PyForward due to conditional compilation
        from persia.prelude import PyForward

        self.dataset = dataset
        self.timeout = timeout
        self.num_workers = num_workers

        if not replica_info:
            current_ctx = cnt_ctx()
            assert current_ctx is not None, "Current conext is None!"
            replica_info = current_ctx.replica_info

        self.forward_engine = PyForward(
            forward_buffer_size, is_training, reproducible, replica_info
        )
        self.forward_engine.set_input_channel(dataset.receiver)
        self.forward_engine.launch(
            torch.cuda.current_device(),
            self.num_workers,
        )

    def __iter__(self):

        for _ in self.dataset:
            yield self.forward_engine.get_batch(self.timeout)

    def __len__(self):
        return len(self.dataset)



#!/usr/bin/env python3

import os
from setuptools import setup, find_packages
from setuptools_rust import Binding, RustExtension


if __name__ == "__main__":
    import colorama

    colorama.init(autoreset=True)

    use_cuda = os.environ.get("USE_CUDA", False)
    features = None if not use_cuda else ["cuda"]

    setup(
        name="persia-core",
        use_scm_version={
            "local_scheme": "no-local-version",
            "root": "..",
            "write_to_template": '__version__ = "{version}"',
            "write_to": os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "python/persia_core/version.py",
            ),
        },
        setup_requires=["setuptools_scm"],
        url="https://github.com/PersiaML/PersiaML/rust",
        python_requires=">=3.7",
        description="Core Python binding for PersiaML.",
        package_dir={"": "python/"},
        packages=find_packages("python/"),
        rust_extensions=[
            RustExtension(
                "persia_core.persia_core",
                path="persia-core/Cargo.toml",
                binding=Binding.PyO3,
                native=True,
                features=features,
            )
        ],
        author="Kuaishou AI Platform",
        author_email="admin@mail.xrlian.com",
        install_requires=[],
        zip_safe=False,
    )



import os

from tqdm import tqdm

from persia.prelude import PyPersiaBatchData
from persia.logger import get_logger
from persia.ctx import BaseCtx
from persia.utils import setup_seed

from data_generator import make_dataloader

logger = get_logger("data_compose")

setup_seed(3)

train_filepath = os.path.join("/data/", "train.npz")

logger.info("init py client done...")
if __name__ == "__main__":
    with BaseCtx() as ctx:
        _, loader = make_dataloader(train_filepath)
        for (dense, batch_sparse_ids, target) in tqdm(loader, desc="gen batch data..."):
            batch_data = PyPersiaBatchData()
            batch_data.add_dense([dense])
            batch_data.add_sparse(batch_sparse_ids)
            batch_data.add_target(target)
            ctx.backend.send_data(batch_data)



import os
import torch


_WORLD_SIZE = int(os.environ.get("WORLD_SIZE", -1))
_REPLICA_SIZE = int(os.environ.get("REPLICA_SIZE", 1))
_REPLICA_INDEX = int(os.environ.get("REPLICA_INDEX", 0))

if torch.cuda.is_available():
    _RANK_ID = int(os.environ.get("RANK", 0))
    _LOCAL_RANK = int(os.environ.get("LOCAL_RANK", 0))
else:
    _RANK_ID = -1
    _LOCAL_RANK = -1


def get_world_size() -> int:
    """Get the total number of processes."""
    return _WORLD_SIZE


def get_rank() -> int:
    """Get the rank of this process."""
    return _RANK_ID


def get_local_rank() -> int:
    """Get the local rank of this process (the rank of the process on the local machine)."""
    return _LOCAL_RANK


def get_replica_size() -> int:
    """Get the replica size of current service.The replica size is the scale number of
    service which launch by docker service or k8s"""
    return _REPLICA_SIZE


def get_replica_index() -> int:
    """Get the replica index of current service.The replica index is the launch order of current service"""

    return _REPLICA_INDEX



import os
import colorama

from setuptools import setup, find_packages
from setuptools_rust import Binding, RustExtension

use_cuda = os.environ.get("USE_CUDA", False)
integration_core = os.environ.get("INTEGRATION_CORE", True)

if __name__ == "__main__":

    colorama.init(autoreset=True)

    extensions = []
    cmdclass = {}

    if use_cuda:
        from torch.utils.cpp_extension import BuildExtension
        from torch.utils.cpp_extension import CUDAExtension

        extensions.append(
            CUDAExtension(
                "persia_torch_ext",
                include_dirs=[
                    os.getcwd() + "/torch_ext/third_party/thread_pool",
                    os.getcwd() + "/torch_ext/third_party/dbg_macro",
                    "/opt/cuda/include",
                ],
                sources=[
                    "torch_ext/persia_torch_ext.cpp",
                ],
                extra_compile_args=["-fopenmp"],
            ),
        )
        cmdclass["build_ext"] = BuildExtension

    rust_extensions = []

    rust_extensions.append(
        RustExtension(
            # TODO: Due to this issue https://github.com/PyO3/setuptools-rust/issues/153 still not release
            # the new version of setuptool_rust, RustExtension can't enable the script feature
            # {
            #     "persia-embedding-sharded-middleware": "persia.persia_middleware",
            #     "persia-embedding-sharded-server": "persia.persia_server"
            # },
            # script=True,
            {
                "persia-embedding-sharded-middleware": "persia.persia-embedding-sharded-middleware",
                "persia-embedding-sharded-server": "persia.persia-embedding-sharded-server",
            },
            path="rust/persia-embedding-sharded-server/Cargo.toml",
            binding=Binding.Exec,
            native=True,
        )
    )

    if integration_core:
        features = None if not use_cuda else ["cuda"]
        rust_extensions.append(
            RustExtension(
                "persia_core",
                path="rust/persia-core/Cargo.toml",
                binding=Binding.PyO3,
                native=True,
                features=features,
            )
        )

    setup(
        name="persia",
        use_scm_version={"local_scheme": "no-local-version"},
        setup_requires=["setuptools_scm"],
        install_requires=["pyyaml", "colorlog", "click"],
        url="https://github.com/PersiaML/PersiaML",
        author="Kuaishou AI Platform Persia Team",
        author_email="admin@mail.xrlian.com",
        description="PersiaML Python Library",
        packages=find_packages(exclude=("tests",)),
        ext_modules=extensions,
        cmdclass=cmdclass,
        rust_extensions=rust_extensions,
        entry_points={
            "console_scripts": [
                "persia_launcher= persia.launcher:cli",
            ]
        },
        python_requires=">=3.7",
    )



import os
import click
import subprocess

from typing import List

from persia.logger import get_logger

_logger = get_logger(__file__)

_DEBUG = int(os.environ.get("DEBUG", False))
_ENV = os.environ.copy()

if _DEBUG:
    # add thirdparty_path into PATH
    persia_dev_path = os.environ.get("PERSIA_DEV_PATH", None)
    if persia_dev_path is not None:
        origin_path = os.environ.get("PATH", "")
        _ENV["PATH"] = f"{persia_dev_path}:{origin_path}"
        _logger.info(f"Use persia_dev_path: {persia_dev_path} to replace origin path")

    # add thirdparty_path into PYTHONPATH
    persia_dev_pythonpath = os.environ.get("PERSIA_DEV_PYTHONPATH", None)
    if persia_dev_pythonpath is not None:
        origin_pythonpath = os.environ.get("PYTHONPATH", "")
        _ENV["PYTHONPATH"] = f"{persia_dev_pythonpath}:{origin_pythonpath}"
        _logger.info(
            f"Use persia_dev_pythonpath: {persia_dev_pythonpath} to replace origin path"
        )


def resolve_binary_execute_path(binary_name: str) -> str:
    if _DEBUG:
        thirdparty_path = os.environ.get("THIRDPARTY_PATH", None)
        if not thirdparty_path:
            raise KeyError(
                "Launch program with debug mode but without THIRDPARTY_PATH env"
            )

        if os.access(os.path.join(thirdparty_path, binary_name), os.X_OK):
            raise Exception(
                f"Can't not found executable {binary_name} in {thirdparty_path}"
            )

        return binary_name
    else:
        return os.path.realpath(os.path.join(__file__, "../", binary_name))


def run_command(cmd: List[str], verb: bool = True):
    cmd = list(map(str, cmd))
    if verb:
        cmd_str = " ".join(cmd)
        print(f"execute command: {cmd_str}")

    subprocess.check_call(cmd, env=_ENV)


@click.group()
def cli():
    ...


@cli.command()
@click.argument("filepath", type=str)
@click.option(
    "--gpu-num", type=int, default=1, help="Number of gpu at current replica node"
)
@click.option("--node-rank", type=int, default=0, help="Replica index of trainer")
@click.option("--nnodes", type=int, default=1, help="Replica num of trainer")
def trainer(filepath, gpu_num: int, node_rank: int, nnodes: int):
    cmd = [
        "python3",
        "-m",
        "torch.distributed.launch",
        "--nproc_per_node",
        gpu_num,
        "--nnodes",
        nnodes,
        "--node_rank",
        node_rank,
        filepath,
    ]
    run_command(cmd)


@cli.command()
@click.argument("filepath", type=str)
@click.option(
    "--replica-index", type=str, default=0, help="Replica index of data compose"
)
@click.option("--replica-size", type=str, default=1, help="Replica num of data compose")
def compose(filepath: str, replica_index: int, replica_size: int):
    cmd = [
        "python3",
        filepath,
        "--replica-index",
        replica_index,
        "--replica-size",
        replica_size,
    ]
    run_command(cmd)


@cli.command()
@click.option("--port", type=int, default=8887, help="Middleware server listen port")
@click.option("--embedding-config", type=str, help="Config of embedding definition")
@click.option(
    "--global-config", type=str, help="Config of embedding server and middleware"
)
@click.option(
    "--replica-index", type=str, default=0, help="Replica index of middleware"
)
@click.option("--replica-size", type=str, default=1, help="Replica num of middleware")
def middleware(
    port: int,
    embedding_config: str,
    global_config: str,
    replica_index: int,
    replica_size: int,
):
    executable_path = resolve_binary_execute_path("persia-embedding-sharded-middleware")
    cmd = [
        executable_path,
        "--port",
        port,
        "--global-config",
        global_config,
        "--embedding-config",
        embedding_config,
        "--replica-index",
        replica_index,
        "--replica-size",
        replica_size,
    ]
    run_command(cmd)


@cli.command()
@click.option("--port", type=int, default=8888, help="Embedding server listen port")
@click.option("--embedding-config", type=str, help="Config of embedding definition")
@click.option(
    "--global-config", type=str, help="Config of embedding server and middleware"
)
@click.option(
    "--replica-index", type=str, default=0, help="Replica index of embedding server"
)
@click.option(
    "--replica-size", type=str, default=1, help="Replica num of embedding server"
)
def server(
    port: int,
    embedding_config: str,
    global_config: str,
    replica_index: int,
    replica_size: int,
):
    executable_path = resolve_binary_execute_path("persia-embedding-sharded-server")
    cmd = [
        executable_path,
        "--port",
        port,
        "--global-config",
        global_config,
        "--embedding-config",
        embedding_config,
        "--shard-idx",
        replica_index,
        "--num-shards",
        replica_size,
    ]
    run_command(cmd)


@cli.command()
def trainer_local():
    # TODO: launch the trainer, middleware server, embedding server
    cmd = "launch the trainer locally with middleware and embedding server"
    print(cmd)


if __name__ == "__main__":
    cli()



from .env import *
from .version import __version__

from persia import logger as logger
from persia import prelude as prelude
from persia import ctx as ctx
from persia import sparse as sparse
from persia import backend as backend
from persia import data as data
from persia import error as error
from persia import service as service
from persia import utils as utils
